from optparse import Values
from pathlib import Path

import numpy as np
import pandas as pd
import pytest
from scipy.sparse import csr_matrix

from ONTraC.niche_net._niche_net import (build_knn_network,
                                         calc_cell_type_composition,
                                         calc_edge_index,
                                         calc_niche_weight_matrix)
from ONTraC.preprocessing.data import load_meta_data

from .utils import temp_dirs


@pytest.fixture
def options() -> Values:
    # Create an options object for testing
    _options = Values()
    _options.meta_input = 'tests/_data/test_data.csv'
    _options.NN_dir = 'tests/_data/NN'
    _options.n_local = 2
    _options.n_neighbors = 5
    return _options


@pytest.fixture()
def sample_data_df(options: Values) -> pd.DataFrame:
    # Create sample data
    sample_data_df = pd.read_csv(options.meta_input)
    sample_data_df = sample_data_df[sample_data_df['Sample'] == 'S1']
    sample_data_df['Cell_Type'] = sample_data_df['Cell_Type'].astype('category')
    return sample_data_df


@pytest.fixture()
def sample_name() -> str:
    return 'S1'


@pytest.fixture()
def dis_matrix() -> np.ndarray:
    return np.array([[0., 1.41421356, 2.82842712, 4.24264069, 5.65685425, 7.07106781],
                     [0., 1.41421356, 1.41421356, 2.82842712, 4.24264069, 5.65685425],
                     [0., 1.41421356, 1.41421356, 2.82842712, 2.82842712, 4.24264069],
                     [0., 1.41421356, 1.41421356, 2.82842712, 2.82842712, 4.24264069],
                     [0., 1.41421356, 1.41421356, 2.82842712, 2.82842712, 4.24264069],
                     [0., 1.41421356, 1.41421356, 2.82842712, 2.82842712, 4.24264069],
                     [0., 1.41421356, 1.41421356, 2.82842712, 4.24264069, 5.65685425],
                     [0., 1.41421356, 2.82842712, 4.24264069, 5.65685425, 7.07106781]])


@pytest.fixture()
def indices_matrix() -> np.ndarray:
    return np.array([[0, 1, 2, 3, 4, 5], [1, 2, 0, 3, 4, 5], [2, 1, 3, 0, 4, 5], [3, 4, 2, 5, 1, 0], [4, 5, 3, 2, 6, 1],
                     [5, 4, 6, 7, 3, 2], [6, 7, 5, 4, 3, 2], [7, 6, 5, 4, 3, 2]])


@pytest.fixture()
def edge_index() -> np.ndarray:
    return np.array([[0, 1], [0, 2], [0, 3], [0, 4], [0, 5], [1, 0], [1, 2], [1, 3], [1, 4], [1, 5], [2, 0], [2, 1],
                     [2, 3], [2, 4], [2, 5], [2, 6], [2, 7], [3, 0], [3, 1], [3, 2], [3, 4], [3, 5], [3, 6], [3, 7],
                     [4, 0], [4, 1], [4, 2], [4, 3], [4, 5], [4, 6], [4, 7], [5, 0], [5, 1], [5, 2], [5, 3], [5, 4],
                     [5, 6], [5, 7], [6, 2], [6, 3], [6, 4], [6, 5], [6, 7], [7, 2], [7, 3], [7, 4], [7, 5], [7, 6]])


@pytest.fixture()
def niche_weight_matrix() -> csr_matrix:
    return csr_matrix([[
        1.00000000e+00, 7.78800783e-01, 3.67879441e-01, 1.05399223e-01, 1.83156386e-02, 1.93045410e-03, 0.00000000e+00,
        0.00000000e+00
    ],
                       [
                           3.67879441e-01, 1.00000000e+00, 3.67879441e-01, 1.83156389e-02, 1.23409799e-04,
                           1.12535168e-07, 0.00000000e+00, 0.00000000e+00
                       ],
                       [
                           1.83156389e-02, 3.67879441e-01, 1.00000000e+00, 3.67879441e-01, 1.83156389e-02,
                           1.23409799e-04, 0.00000000e+00, 0.00000000e+00
                       ],
                       [
                           1.23409799e-04, 1.83156389e-02, 3.67879441e-01, 1.00000000e+00, 3.67879441e-01,
                           1.83156389e-02, 0.00000000e+00, 0.00000000e+00
                       ],
                       [
                           0.00000000e+00, 1.23409799e-04, 1.83156389e-02, 3.67879441e-01, 1.00000000e+00,
                           3.67879441e-01, 1.83156389e-02, 0.00000000e+00
                       ],
                       [
                           0.00000000e+00, 0.00000000e+00, 1.23409799e-04, 1.83156389e-02, 3.67879441e-01,
                           1.00000000e+00, 3.67879441e-01, 1.83156389e-02
                       ],
                       [
                           0.00000000e+00, 0.00000000e+00, 1.12535168e-07, 1.23409799e-04, 1.83156389e-02,
                           3.67879441e-01, 1.00000000e+00, 3.67879441e-01
                       ],
                       [
                           0.00000000e+00, 0.00000000e+00, 1.93045410e-03, 1.83156386e-02, 1.05399223e-01,
                           3.67879441e-01, 7.78800783e-01, 1.00000000e+00
                       ]])


@pytest.fixture()
def cell_type_composition() -> np.ndarray:
    return np.array([[0.61003367, 0.38996633], [0.41949784, 0.58050216], [0.58483686, 0.41516314],
                     [0.41516314, 0.58483686], [0.58483686, 0.41516314], [0.41516314, 0.58483686],
                     [0.58050216, 0.41949784], [0.38996633, 0.61003367]])


def test_load_meta_data(options: Values) -> None:
    """
    Test the load_meta_data module.
    :param options: Values, options.
    :return: None.
    """

    with temp_dirs(options=options):
        # load_meta_data function should:
        # 1) read original data file (options.meta_input)
        # 2) retrun original data with Cell_ID, Sample, Cell_Type, x, and y columns
        # 3) Cell_ID should be unique
        # 4) Cell_Type should be categorical
        # 5) save `cell_type_code.csv` file in the preprocessing directory
        ori_data_df = load_meta_data(save_dir=options.NN_dir, meta_data_file=options.meta_input)

        # Check the expected output data shape
        assert ori_data_df.shape[0] == 16  # Check if the DataFrame shape is unchanged

        # Check if the expected columns are present
        assert 'Cell_ID' in ori_data_df.columns
        assert 'Sample' in ori_data_df.columns
        assert 'Cell_Type' in ori_data_df.columns
        assert 'x' in ori_data_df.columns
        assert 'y' in ori_data_df.columns

        # Check if Cell_ID is unique
        # TODO: need an duplicate Cell_ID in different samples case here
        # assert not ori_data_df['Cell_ID'].duplicated().any()

        # Check if Cell_Type is categorical
        assert ori_data_df['Cell_Type'].dtype.name == 'category'

        # Check if the `cell_type_code.csv` file is saved
        assert Path(f'{options.NN_dir}/cell_type_code.csv').exists()

        # Check the content of the `cell_type_code.csv` file
        gen_cell_type_code = pd.read_csv(f'{options.NN_dir}/cell_type_code.csv')
        assert gen_cell_type_code.equals(pd.DataFrame({'Code': [0, 1], 'Cell_Type': ['A', 'B']}))


def test_build_knn_network(options: Values, sample_data_df: pd.DataFrame, sample_name: str, dis_matrix: np.ndarray,
                           indices_matrix: np.ndarray) -> None:
    """
    Test the build_knn_network function.
    :param options: Values, options.
    :param sample_data_df: pd.DataFrame, sample data.
    :param sample_name: str, sample name.
    :param dis_matrix: np.ndarray, distance matrix.
    :param indices_matrix: np.ndarray, indices matrix.
    :return: None.
    """

    # Call the function
    gen_coordinates, gen_dis_matrix, gen_indices_matrix = build_knn_network(sample_name=sample_name,
                                                                            sample_meta_df=sample_data_df,
                                                                            n_neighbors=options.n_neighbors,
                                                                            n_local=options.n_local)

    # Check the output types
    assert isinstance(gen_coordinates, np.ndarray)
    assert isinstance(gen_dis_matrix, np.ndarray)
    assert isinstance(gen_indices_matrix, np.ndarray)

    # Check the shape of the output arrays
    assert gen_coordinates.shape == (8, 2)
    assert gen_dis_matrix.shape == (8, 6)
    assert gen_indices_matrix.shape == (8, 6)

    # Check if the coordinates are correct
    assert np.array_equal(gen_coordinates, np.array(sample_data_df[['x', 'y']].values))

    # Check if the distance matrix is correct
    assert np.allclose(a=gen_dis_matrix, b=dis_matrix, atol=1e-7)

    # Check if the indices matrix is correct
    assert np.array_equal(a1=gen_indices_matrix, a2=indices_matrix)


def test_calc_edge_index(options: Values, sample_data_df: pd.DataFrame, sample_name: str, indices_matrix: np.ndarray,
                         edge_index: np.ndarray) -> None:
    """
    Test the calc_edge_index function.
    :param options: Values, options.
    :param sample_data_df: pd.DataFrame, sample data.
    :param sample_name: str, sample name.
    :param indices_matrix: np.ndarray, indices matrix.
    :param edge_index: np.ndarray, edge index.
    :return: None.
    """

    # Call the function
    gen_edge_index = calc_edge_index(sample_name=sample_name,
                                     sample_meta_df=sample_data_df,
                                     indices_matrix=indices_matrix,
                                     n_neighbors=options.n_neighbors)

    # Check the output type
    assert isinstance(gen_edge_index, np.ndarray)

    # Check the shape of the output array
    assert gen_edge_index.shape == (48, 2)

    # Check if the edge index is correct
    assert np.array_equal(gen_edge_index, edge_index)


def test_calc_niche_weight_matrix(options: Values, sample_data_df: pd.DataFrame, dis_matrix: np.ndarray,
                                  indices_matrix: np.ndarray, niche_weight_matrix: csr_matrix) -> None:
    """
    Test the calc_niche_weight_matrix function.
    :param options: Values, options.
    :param sample_data_df: pd.DataFrame, sample data.
    :param dis_matrix: np.ndarray, distance matrix.
    :param indices_matrix: np.ndarray, indices matrix.
    :param niche_weight_matrix: csr_matrix, niche weight matrix.
    :return: None.
    """

    # Call the function
    gen_niche_weight_matrix = calc_niche_weight_matrix(sample_name='sample',
                                                       sample_meta_df=sample_data_df,
                                                       dis_matrix=dis_matrix,
                                                       indices_matrix=indices_matrix,
                                                       n_neighbors=options.n_neighbors,
                                                       n_local=options.n_local)

    # Check the output type
    assert isinstance(gen_niche_weight_matrix, csr_matrix)

    # Check the shape of the output matrix
    assert gen_niche_weight_matrix.shape == (8, 8)

    # Check if the niche weight matrix is correct
    assert np.allclose(a=gen_niche_weight_matrix.toarray(), b=niche_weight_matrix.toarray(), atol=1e-7)


def test_calc_cell_type_composition(sample_data_df: pd.DataFrame, niche_weight_matrix: csr_matrix,
                                    cell_type_composition: np.ndarray) -> None:
    """
    Test the calc_cell_type_composition function.
    :param options: Values, options.
    :param sample_data_df: pd.DataFrame, sample data.
    :param niche_weight_matrix: csr_matrix, niche weight matrix.
    :param cell_type_composition: np.ndarray, cell type composition.
    :return: None.
    """

    # Call the function
    gen_cell_type_composition = calc_cell_type_composition(sample_name='sample',
                                                           sample_meta_df=sample_data_df,
                                                           niche_weight_matrix=niche_weight_matrix,
                                                           decompsited_cell_type=None)

    # Check the output type
    assert isinstance(gen_cell_type_composition, np.ndarray)

    # Check the shape of the output array
    assert gen_cell_type_composition.shape == (8, 2)

    # Check if the cell type composition is correct
    assert np.allclose(a=gen_cell_type_composition, b=cell_type_composition, atol=1e-7)



from typing import Optional, Tuple

import torch
from torch import Tensor
from torch import nn
from copy import deepcopy

from ..log import *
from .dmon_exp_pool import DMoNPooling
from .norm_dense_gcn_conv import NormDenseGCNConv


class GraphPooling(torch.nn.Module):
    """
    GraphPooling
    """

    def __init__(self, input_feats, k: int, dropout: float = 0, exponent: float = 1, *args, **kwargs) -> None:
        super().__init__(*args, **kwargs)
        self.dropout = dropout
        self.exponent = exponent
        self.pool = DMoNPooling(channels=input_feats, k=k, dropout=0, exponent=self.exponent)
        self.k = k

        self.reset_parameters()

    def reset_parameters(self) -> None:
        self.pool.reset_parameters()

    def forward(self,
                x: Tensor,
                adj: Tensor,
                mask: Optional[Tensor] = None) -> Tuple[Tensor, Tensor, Tensor, Tensor, Tensor, Tensor]:
        r"""
        forward function
        Args:
            x (torch.Tensor): Node feature tensor
                :math:`\mathbf{X} \in \mathbb{R}^{B \times N \times F}`, with
                batch-size :math:`B`, (maximum) number of nodes :math:`N` for
                each graph, and feature dimension :math:`F`.
            adj (torch.Tensor): Adjacency tensor
                :math:`\mathbf{A} \in \mathbb{R}^{B \times N \times N}`.
            mask (torch.Tensor, optional): Mask matrix
                :math:`\mathbf{M} \in {\{ 0, 1 \}}^{B \times N}` indicating
                the valid nodes for each graph. (default: :obj:`None`)
        Returns:
            Tensor: output feature matrix
        """
        s, out, out_adj, spectral_loss, ortho_loss, cluster_loss = self.pool(x=x, adj=adj, mask=mask)
        return s, out, out_adj, spectral_loss, ortho_loss, cluster_loss

    def predict(self, x: Tensor, adj: Tensor, mask: Optional[Tensor] = None) -> Tuple[Tensor, Tensor, Tensor]:
        s, out, out_adj, *_ = self.pool(x=x, adj=adj, mask=mask)
        return s, out, out_adj


class GNN(torch.nn.Module):
    """
    GCN + GraphPooling
    """

    def __init__(self,
                 input_feats: int,
                 hidden_feats: int,
                 k: int,
                 n_gcn_layers: int = 2,
                 dropout: float = 0,
                 exponent: float = 1,
                 *args,
                 **kwargs) -> None:
        super().__init__(*args, **kwargs)
        self.n_gcn_layers = n_gcn_layers
        self.gcns = nn.ModuleList([NormDenseGCNConv(input_feats if i == 0 else hidden_feats, hidden_feats)
                                   for i in range(self.n_gcn_layers)])
        self.activations = nn.ModuleList([torch.nn.SELU() for _ in range(self.n_gcn_layers)])
        self.pool = GraphPooling(input_feats=hidden_feats, k=k, dropout=dropout, exponent=exponent)
        self.k = k

        self.reset_parameters()

    def reset_parameters(self) -> None:
        for gcn in self.gcns:
            gcn.reset_parameters()
        self.pool.reset_parameters()

    def forward(self,
                x: Tensor,
                adj: Tensor,
                mask: Optional[Tensor] = None) -> Tuple[Tensor, Tensor, Tensor, Tensor, Tensor, Tensor]:
        r"""
        forward function
        X' = \mathbf{\hat{L}}X\mathbf{\Theta}
        \mathbf{\hat{A}} = \mathbf{A} + \mathbf{I}
        \mathbf(\hat{D}) = \sum_{j=1}^N \mathbf{\hat{A}}_{ij}
        \mathbf{\hat{L}} = \mathbf{\hat{D}}^{-1/2}\mathbf{\hat{A}}\mathbf{\hat{D}}^{-1/2}
        Args:
            x (torch.Tensor): Node feature tensor
                :math:`\mathbf{X} \in \mathbb{R}^{B \times N \times F}`, with
                batch-size :math:`B`, (maximum) number of nodes :math:`N` for
                each graph, and feature dimension :math:`F`.
            adj (torch.Tensor): Adjacency tensor
                :math:`\mathbf{A} \in \mathbb{R}^{B \times N \times N}`.
            mask (torch.Tensor, optional): Mask matrix
                :math:`\mathbf{M} \in {\{ 0, 1 \}}^{B \times N}` indicating
                the valid nodes for each graph. (default: :obj:`None`)
        Returns:
            Tensor: output feature matrix
        """
        for i in range(self.n_gcn_layers):
            x = self.activations[i](self.gcns[i](x=x, adj=adj, mask=mask))
        s, out, out_adj, spectral_loss, ortho_loss, cluster_loss = self.pool(x=x, adj=adj, mask=mask)
        return s, out, out_adj, spectral_loss, ortho_loss, cluster_loss

    def evaluate(self,
                 x: Tensor,
                 adj: Tensor,
                 mask: Optional[Tensor] = None) -> Tuple[Tensor, Tensor, Tensor, Tensor, Tensor, Tensor]:
        
        for i in range(self.n_gcn_layers):
            x = self.activations[i](self.gcns[i](x=x, adj=adj, mask=mask))
        s, out, out_adj, spectral_loss, ortho_loss, cluster_loss = self.pool(x=x, adj=adj, mask=mask)
        return s, out, out_adj, spectral_loss, ortho_loss, cluster_loss

    def predict(self, x: Tensor, adj: Tensor, mask: Optional[Tensor] = None) -> Tuple[Tensor, Tensor, Tensor]:
        r"""
        predict function
        Args:
            x (torch.Tensor): Node feature tensor
                :math:`\mathbf{X} \in \mathbb{R}^{B \times N \times F}`, with
                batch-size :math:`B`, (maximum) number of nodes :math:`N` for
                each graph, and feature dimension :math:`F`.
            adj (torch.Tensor): Adjacency tensor
                :math:`\mathbf{A} \in \mathbb{R}^{B \times N \times N}`.
            mask (torch.Tensor, optional): Mask matrix\
                :math:`\mathbf{M} \in {\{ 0, 1 \}}^{B \times N}` indicating\
                the valid nodes for each graph. (default: :obj:`None`)
        Returns:
            s (torch.Tensor): Node assignment matrix
                :math:`\mathbf{S} \in \mathbb{R}^{B \times N \times K}`
            out (torch.Tensor): Output feature matrix
                :math:`\mathbf{X} \in \mathbb{R}^{B \times K \times H}`
            out_adj (torch.Tensor): Output adjacency matrix
                :math:`\mathbf{A} \in \mathbb{R}^{B \times K \times K}`
            """
        
        for i in range(self.n_gcn_layers):
            x = self.activations[i](self.gcns[i](x=x, adj=adj, mask=mask))
        s, out, out_adj, *_ = self.pool(x=x, adj=adj, mask=mask)
        return s, out, out_adj

    def predict_embed(self, x: Tensor, adj: Tensor, mask: Optional[Tensor] = None) -> Tensor:
        for i in range(self.n_gcn_layers):
            x = self.activations[i](self.gcns[i](x=x, adj=adj, mask=mask))
        return x


__all__ = ['GraphPooling', 'GNN']



import random
from pathlib import Path
from typing import Callable, Dict, List, Optional, Tuple, Type, Union

import numpy as np
import pandas as pd
import torch
from scipy.sparse import load_npz
from torch import Tensor
from torch_geometric.loader import DenseDataLoader

from ..data import SpatailOmicsDataset
from ..log import info
from ..train import SubBatchTrainProtocol


def set_seed(seed: int) -> None:
    """
    Set seed.
    :param seed: seed.
    :return: None.
    """

    random.seed(seed)
    torch.manual_seed(seed)
    np.random.seed(seed)


def train(nn_model: torch.nn.Module,
          BatchTrain: Type[SubBatchTrainProtocol],
          sample_loader: DenseDataLoader,
          device: torch.device,
          max_epochs: int,
          max_patience: int,
          min_delta: float,
          min_epochs: int,
          lr: float,
          save_dir: Union[str, Path],
          inspect_funcs: Optional[List[Callable]] = None,
          **kwargs) -> SubBatchTrainProtocol:
    """
    GNN training process.
    :param nn_model: nn model.
    :param BatchTrain: Type[SubBatchTrainProtocol], batch train.
    :param sample_loader: DenseDataLoader, sample loader.
    :param device: torch.device, device.
    :param max_epochs: int, max epochs.
    :param max_patience: int, max patience.
    :param min_delta: float, min delta.
    :param min_epochs: int, min epochs.
    :param lr: float, learning rate.
    :param save_dir: Union[str, Path], save directory.
    :param inspect_funcs: Optional[List[Callable]], inspect functions.
    :param kwargs: dict, loss weight arguments.
    """
    optimizer = torch.optim.Adam(nn_model.parameters(), lr=lr)
    batch_train = BatchTrain(model=nn_model, device=torch.device(device), data_loader=sample_loader)  # type: ignore
    batch_train.save(path=f'{save_dir}/epoch_0.pt')

    loss_weight_args: Dict[str, float] = {key: value for key, value in kwargs.items() if key.endswith('loss_weight')}

    batch_train.train(optimizer=optimizer,
                      inspect_funcs=inspect_funcs,
                      max_epochs=max_epochs,
                      max_patience=max_patience,
                      min_delta=min_delta,
                      min_epochs=min_epochs,
                      output=save_dir,
                      **loss_weight_args)
    batch_train.save(path=f'{save_dir}/model_state_dict.pt')
    info(message=f'Training process end.')
    return batch_train


def evaluate(batch_train: SubBatchTrainProtocol) -> None:
    """
    Evaluate the performance of ONTraC model on data.
    :param batch_train: SubBatchTrainProtocol, batch train.
    :return
    """
    info(message=f'Evaluating process start.')
    loss_dict: Dict[str, np.floating] = batch_train.evaluate()  # type: ignore
    info(message=f'Evaluate loss, {repr(loss_dict)}')
    info(message=f'Evaluating process end.')


def predict(output_dir: str, batch_train: SubBatchTrainProtocol,
            dataset: SpatailOmicsDataset) -> Tuple[Optional[np.ndarray], Optional[np.ndarray]]:
    """
    Predict the results of ONTraC model on data.
    :param output_dir: str, output directory.
    :param batch_train: SubBatchTrainProtocol, batch train.
    :param dataset: SpatailOmicsDataset, dataset.
    :return: consolidate_s_array, consolidate_out_adj_array.
    """
    info(f'Predicting process start.')
    each_sample_loader = DenseDataLoader(dataset, batch_size=1)
    consolidate_flag = False
    consolidate_s_list = []
    consolidate_out = None
    consolidate_out_adj = None
    for data in each_sample_loader:  # type: ignore
        info(f'Generating prediction results for {data.name[0]}.')
        data = data.to(batch_train.device)  # type: ignore
        predict_result = batch_train.predict_dict(data=data)  # type: ignore
        for key, value in predict_result.items():
            np.savetxt(fname=f'{output_dir}/{data.name[0]}_{key}.csv.gz',
                       X=value.squeeze(0).detach().cpu().numpy(),
                       delimiter=',')

        # consolidate results
        if not consolidate_flag and ('s' in predict_result and 'out' in predict_result and 'out_adj' in predict_result):
            consolidate_flag = True
        if consolidate_flag:
            s = predict_result['s']
            out = predict_result['out']
            s = s.squeeze(0)
            consolidate_s_list.append(s)
            out_adj_ = torch.matmul(torch.matmul(s.T, data.adj.squeeze(0)), s)
            consolidate_out_adj: Tensor = out_adj_ if consolidate_out_adj is None else consolidate_out_adj + out_adj_
            consolidate_out: Tensor = out.squeeze(
                0) * data.mask.sum() if consolidate_out is None else consolidate_out + out.squeeze(0) * data.mask.sum()

    consolidate_s_array, consolidate_out_adj_array = None, None
    if consolidate_flag:
        # consolidate out
        nodes_num = 0
        for data in each_sample_loader:  # type: ignore
            nodes_num += data.mask.sum()
        consolidate_out = consolidate_out / nodes_num  # type: ignore
        consolidate_out_array = consolidate_out.detach().cpu().numpy()  # type: ignore
        np.savetxt(fname=f'{output_dir}/consolidate_out.csv.gz', X=consolidate_out_array, delimiter=',')
        # consolidate s
        consolidate_s = torch.cat(consolidate_s_list, dim=0)
        # consolidate out_adj
        ind = torch.arange(consolidate_s.shape[-1], device=consolidate_out_adj.device)  # type: ignore
        consolidate_out_adj[ind, ind] = 0  # type: ignore
        d = torch.einsum('ij->i', consolidate_out_adj)
        d = torch.sqrt(d)[:, None] + 1e-15
        consolidate_out_adj = (consolidate_out_adj / d) / d.transpose(0, 1)
        consolidate_s_array = consolidate_s.detach().cpu().numpy()
        consolidate_out_adj_array = consolidate_out_adj.detach().cpu().numpy()
        np.savetxt(fname=f'{output_dir}/consolidate_s.csv.gz', X=consolidate_s_array, delimiter=',')
        np.savetxt(fname=f'{output_dir}/consolidate_out_adj.csv.gz', X=consolidate_out_adj_array, delimiter=',')

    info(f'Predicting process end.')
    return consolidate_s_array, consolidate_out_adj_array


def save_graph_pooling_results(meta_data_df: pd.DataFrame, dataset: SpatailOmicsDataset, rel_params: Dict,
                               consolidate_s_array: np.ndarray, output_dir: str) -> None:
    """
    Save graph pooling results as the Niche cluster (max probability for each niche & cell).
    :param meta_data_df: pd.DataFrame, original data. Sample and Cell_ID columns are used.
    :param dataset: SpatailOmicsDataset, dataset.
    :param rel_params: dict, relative parameters.
    :param consolidate_s_array: np.ndarray, consolidate s array.
    :param output_dir: str, output directory.
    :return: None.
    """

    consolidate_s_niche_df = pd.DataFrame()
    consolidate_s_cell_df = pd.DataFrame()
    for i, data in enumerate(dataset):
        # the slice of data in each sample
        slice_ = slice(i * data.x.shape[0], i * data.x.shape[0] + data.mask.sum())
        consolidate_s = consolidate_s_array[slice_]  # N x C
        consolidate_s_df_ = pd.DataFrame(consolidate_s,
                                         columns=[f'NicheCluster_{i}' for i in range(consolidate_s.shape[1])])
        consolidate_s_df_['Cell_ID'] = meta_data_df[meta_data_df['Sample'] == data.name]['Cell_ID'].values
        consolidate_s_niche_df = pd.concat([consolidate_s_niche_df, consolidate_s_df_], axis=0)

        # niche to cell matrix
        niche_weight_matrix = load_npz(rel_params['Data'][i]['NicheWeightMatrix'])
        niche_to_cell_matrix = (
            niche_weight_matrix /
            niche_weight_matrix.sum(axis=0)).T  # normalize by the all niches associated with each cell, N x N

        consolidate_s_cell = niche_to_cell_matrix @ consolidate_s
        consolidate_s_cell_df_ = pd.DataFrame(consolidate_s_cell,
                                              columns=[f'NicheCluster_{i}' for i in range(consolidate_s_cell.shape[1])])
        consolidate_s_cell_df_['Cell_ID'] = meta_data_df[meta_data_df['Sample'] == data.name]['Cell_ID'].values
        consolidate_s_cell_df = pd.concat([consolidate_s_cell_df, consolidate_s_cell_df_], axis=0)

    consolidate_s_niche_df = consolidate_s_niche_df.set_index('Cell_ID')
    consolidate_s_niche_df = consolidate_s_niche_df.loc[meta_data_df['Cell_ID'], :]
    consolidate_s_niche_df.to_csv(f'{output_dir}/niche_level_niche_cluster.csv.gz',
                                  index=True,
                                  index_label='Cell_ID',
                                  header=True)
    consolidate_s_niche_df['Niche_Cluster'] = consolidate_s_niche_df.values.argmax(axis=1)
    consolidate_s_niche_df['Niche_Cluster'].to_csv(f'{output_dir}/niche_level_max_niche_cluster.csv.gz',
                                                   index=True,
                                                   header=True)
    consolidate_s_cell_df = consolidate_s_cell_df.set_index('Cell_ID')
    consolidate_s_cell_df = consolidate_s_cell_df.loc[meta_data_df['Cell_ID'], :]
    consolidate_s_cell_df.to_csv(f'{output_dir}/cell_level_niche_cluster.csv.gz',
                                 index=True,
                                 index_label='Cell_ID',
                                 header=True)
    consolidate_s_cell_df['Niche_Cluster'] = consolidate_s_cell_df.values.argmax(axis=1)
    consolidate_s_cell_df['Niche_Cluster'].to_csv(f'{output_dir}/cell_level_max_niche_cluster.csv.gz',
                                                  index=True,
                                                  header=True)



# pre-processing logic
# option 1: niche network construction
#   1) load meta data
#   2) generate ct code
#       - if cell-level data, generate ct code directly from meta data
#       - if spot-level data, load low res exp data, generate spotXct matrix, and generate ct code
# option 2: load data from NN-dir




import os
import shutil
from optparse import Values
from typing import List, Optional

import pandas as pd

from ..log import *
from ..optparser._IO import write_io_options_memo
from ..optparser._NN import write_niche_net_constr_memo
from ..optparser._NT import write_NT_options_memo
from ..optparser._train import write_GCN_options_memo, write_GP_options_memo, write_train_options_memo
from ..run.processes import niche_trajectory_construct, gnn, niche_network_construct


def io_opt_valid(options: Values, process='ontrac', io_options: Optional[List[str]] = None) -> Values:
    """
    Validate I/O options
    :param options: options
    :param process: str, process name
    :param io_options: List of I/O options.
    :return: options
    """

    # TODO: valida based on process

    if io_options is None:
        return options

    # input files
    if 'input' in io_options:

        # meta_input
        if hasattr(options, 'dataset') and not hasattr(options, 'meta_input'):
            warning('The dataset option will be deprecated from v3.0. Please use meta_input instead.')
            options.meta_input = options.dataset
        if not hasattr(options, 'meta_input'):
            raise AttributeError('Please provide a meta_input in options.')
        if not os.path.isfile(options.meta_input):
            raise ValueError(f'The input file ({options.meta_input}) you given does not exist.')
        if not options.meta_input.endswith(('csv', 'csv.gz')):
            raise ValueError(f'The input file ({options.meta_input}) should be in csv format.')

        # low resolution expression data
        if hasattr(options, 'deconvoluted_ct_composition'):
            if not os.path.isfile(options.deconvoluted_ct_composition):
                raise ValueError(f'The The deconvoluted cell type composition file ({options.deconvoluted_ct_composition}) you given does not exist.')
            if not options.deconvoluted_ct_composition.endswith(('csv', 'csv.gz')):
                raise ValueError(f'The deconvoluted cell type composition file ({options.deconvoluted_ct_composition}) should be in csv format.')

    # NN_dir
    if 'NN_dir' in io_options:
        if hasattr(options, 'preprocessing_dir') and not hasattr(options, 'NN_dir'):
            warning('The preprocessing_dir option will be deprecated from v3.0. Please use NN_dir instead.')
            options.NN_dir = options.preprocessing_dir
        if not hasattr(options, 'NN_dir'):
            raise AttributeError('Please provide a directory for niche network outputs.')
        if os.path.isdir(options.NN_dir):
            warning(f'The directory ({options.NN_dir}) you given already exists. It will be overwritten.')
            shutil.rmtree(options.NN_dir)
        else:
            info(f'Create directory: {options.NN_dir}')
        os.makedirs(options.NN_dir)

    # GNN_dir
    if 'GNN_dir' in io_options:
        if not hasattr(options, 'GNN_dir'):
            raise AttributeError('Please provide a directory for the GNN output.')
        if os.path.isdir(options.GNN_dir):
            warning(f'The directory ({options.GNN_dir}) you given already exists. It will be overwritten.')
            shutil.rmtree(options.GNN_dir)
        else:
            info(f'Create directory: {options.GNN_dir}')
        os.makedirs(options.GNN_dir)

    # NT_dir
    if 'NT_dir' in io_options:
        if hasattr(options, 'NTScore_dir') and not hasattr(options, 'NT_dir'):
            warning('The NTScore_dir option will be deprecated from v3.0. Please use NT_dir instead.')
            options.NT_dir = options.NTScore_dir
        if not hasattr(options, 'NT_dir'):
            raise AttributeError('Please provide a directory for the niche trajectory output.')
        if os.path.isdir(options.NT_dir):
            warning(f'The directory ({options.NT_dir}) you given already exists. It will be overwritten.')
            shutil.rmtree(options.NT_dir)
        else:
            info(f'Create directory: {options.NT_dir}')
        os.makedirs(options.NT_dir)

    return options


def preprocessing_opt_valid(options: Values, process='ontrac') -> Values:
    """
    Validate preprocessing options
    :param options: options
    :param process: str, process name
    :return: options
    """

    return options


def niche_net_opt_valid(options: Values, process='ontrac') -> Values:
    """
    Validate niche network construction options
    :param options: options
    :param process: str, process name
    :return: options
    """

    # TODO: valida based on process

    # n_cpu
    if not hasattr(options, 'n_cpu'):
        options.n_cpu = 4
    elif not isinstance(options.n_cpu, int):
        raise ValueError(f'n_cpu should be an integer. You provided {options.n_cpu}.')
    else:
        options.n_cpu = int(options.n_cpu)
    if options.n_cpu < 1:
        raise ValueError(f'n_cpu should be greater than 0. You provided {options.n_cpu}.')

    # n_neighbors
    if not hasattr(options, 'n_neighbors'):
        options.n_neighbors = 50
    elif not isinstance(options.n_neighbors, int):
        raise ValueError(f'n_neighbors should be an integer. You provided {options.n_neighbors}.')
    else:
        options.n_neighbors = int(options.n_neighbors)
    if options.n_neighbors < 1:
        raise ValueError(f'n_neighbors should be greater than 0. You provided {options.n_neighbors}.')

    # n_local
    if not hasattr(options, 'n_local'):
        options.n_local = 20
    elif not isinstance(options.n_local, int):
        raise ValueError(f'n_local should be an integer. You provided {options.n_local}.')
    else:
        options.n_local = int(options.n_local)
    if options.n_local < 2:
        raise ValueError(f'n_local should be greater than 1. You provided {options.n_local}.')

    return options


def gnn_opt_valid(options: Values, process='ontrac') -> Values:
    """
    Validate GNN options
    :param options: options
    :param process: str, process name
    :return: options
    """

    # TODO: valida based on process

    # device
    if not hasattr(options, 'device'):
        options.device = 'cpu'
    elif not isinstance(options.device, str):
        raise ValueError(f'device should be a string. You provided {options.device}.')
    else:
        options.device = str(options.device)
    if options.device not in ['cpu', 'cuda']:
        raise ValueError(f'device should be either "cpu" or "cuda". You provided {options.device}.')

    # epochs
    if not hasattr(options, 'epochs'):
        options.epochs = 1000
    elif not isinstance(options.epochs, int):
        raise ValueError(f'epochs should be an integer. You provided {options.epochs}.')
    else:
        options.epochs = int(options.epochs)
    if options.epochs < 1:
        raise ValueError(f'epochs should be greater than 0. You provided {options.epochs}.')

    # patience
    if not hasattr(options, 'patience'):
        options.patience = 100
    elif not isinstance(options.patience, int):
        raise ValueError(f'patience should be an integer. You provided {options.patience}.')
    else:
        options.patience = int(options.patience)
    if options.patience < 1:
        raise ValueError(f'patience should be greater than 0. You provided {options.patience}.')

    # min_delta
    if not hasattr(options, 'min_delta'):
        options.min_delta = 0.001
    elif not isinstance(options.min_delta, float):
        raise ValueError(f'min_delta should be a float. You provided {options.min_delta}.')
    else:
        options.min_delta = float(options.min_delta)
    if not 0 < options.min_delta < 1:
        raise ValueError(f'min_delta should be in the range of (0, 1). You provided {options.min_delta}.')

    # min_epochs
    if not hasattr(options, 'min_epochs'):
        options.min_epochs = 50
    elif not isinstance(options.min_epochs, int):
        raise ValueError(f'min_epochs should be an integer. You provided {options.min_epochs}.')
    else:
        options.min_epochs = int(options.min_epochs)
    if options.min_epochs < 1:
        raise ValueError(f'min_epochs should be greater than 0. You provided {options.min_epochs}.')

    # batch_size
    if not hasattr(options, 'batch_size'):
        options.batch_size = 0
    elif not isinstance(options.batch_size, int):
        raise ValueError(f'batch_size should be an integer. You provided {options.batch_size}.')
    else:
        options.batch_size = int(options.batch_size)
    if options.batch_size < 0:
        raise ValueError(f'batch_size should be greater than or equal to 0. You provided {options.batch_size}.')

    # seed
    if not hasattr(options, 'seed'):
        options.seed = 42
    elif not isinstance(options.seed, int):
        raise ValueError(f'seed should be an integer. You provided {options.seed}.')
    else:
        options.seed = int(options.seed)

    # lr
    if not hasattr(options, 'lr'):
        options.lr = 0.03
    elif not isinstance(options.lr, float):
        raise ValueError(f'Learning rate should be a float. You provided {options.lr}.')
    else:
        options.lr = float(options.lr)
    if options.lr < 0:
        raise ValueError(f'Learning rate should be greater than 0. You provided {options.lr}.')

    # hidden_feats
    if not hasattr(options, 'hidden_feats'):
        options.hidden_feats = 4
    elif not isinstance(options.hidden_feats, int):
        raise ValueError(f'hidden_feats should be an integer. You provided {options.hidden_feats}.')
    else:
        options.hidden_feats = int(options.hidden_feats)
    if options.hidden_feats < 1:
        raise ValueError(f'hidden_feats should be greater than 0. You provided {options.hidden_feats}.')

    # n_gcn_layers
    if not hasattr(options, 'n_gcn_layers'):
        options.n_gcn_layers = 2
    elif not isinstance(options.n_gcn_layers, int):
        raise ValueError(f'n_gcn_layers should be an integer. You provided {options.n_gcn_layers}.')
    else:
        options.n_gcn_layers = int(options.n_gcn_layers)

    # k
    if not hasattr(options, 'k'):
        options.k = 6
    elif not isinstance(options.k, int):
        raise ValueError(f'k-cluster should be an integer. You provided {options.k}.')
    else:
        options.k = int(options.k)
    if options.k < 2:
        raise ValueError(f'k-cluster should be greater than 1. You provided {options.k}.')

    # modularity_loss_weight
    if not hasattr(options, 'modularity_loss_weight'):
        options.modularity_loss_weight = 0.3
    elif not isinstance(options.modularity_loss_weight, float):
        raise ValueError(f'modularity_loss_weight should be a float. You provided {options.modularity_loss_weight}.')
    else:
        options.modularity_loss_weight = float(options.modularity_loss_weight)
    if options.modularity_loss_weight < 0:
        raise ValueError(
            f'modularity_loss_weight should be greater than or equal to 0. You provided {options.modularity_loss_weight}.'
        )

    # purity_loss_weight
    if not hasattr(options, 'purity_loss_weight'):
        options.purity_loss_weight = 300.0
    elif not isinstance(options.purity_loss_weight, float):
        raise ValueError(f'purity_loss_weight should be a float. You provided {options.purity_loss_weight}.')
    else:
        options.purity_loss_weight = float(options.purity_loss_weight)
    if options.purity_loss_weight < 0:
        raise ValueError(
            f'purity_loss_weight should be greater than or equal to 0. You provided {options.purity_loss_weight}.')

    # regularization_loss_weight
    if not hasattr(options, 'regularization_loss_weight'):
        options.regularization_loss_weight = 0.1
    elif not isinstance(options.regularization_loss_weight, float):
        raise ValueError(
            f'regularization_loss_weight should be a float. You provided {options.regularization_loss_weight}.')
    else:
        options.regularization_loss_weight = float(options.regularization_loss_weight)
    if options.regularization_loss_weight < 0:
        raise ValueError(
            f'regularization_loss_weight should be greater than or equal to 0. You provided {options.regularization_loss_weight}.'
        )

    # beta
    if not hasattr(options, 'beta'):
        options.beta = 0.03
    elif not isinstance(options.beta, float):
        raise ValueError(f'beta should be a float. You provided {options.beta}.')
    else:
        options.beta = float(options.beta)
    if options.beta < 0:
        raise ValueError(f'beta should be greater than or equal to 0. You provided {options.beta}.')

    return options


def nt_opt_valid(options: Values, process='ontrac') -> Values:
    """
    Validate niche trajectory options.
    Placehold and do nothing.

    :param options: options
    :param process: str, process name
    :return: options
    """

    return options


def options_valid(options: Values, process='ontrac') -> Values:
    """
    Validate options
    :param options: options
    :param process: str, process name
    :return: options
    """

    # TODO: valida based on process
    # TODO: integrated into optparser module

    io_options = ['input', 'NN_dir', 'GNN_dir', 'NT_dir']

    # I/O options
    options = io_opt_valid(options=options, process=process, io_options=io_options)
    # preprocessing
    options = preprocessing_opt_valid(options=options, process=process)
    # niche network construction options
    options = niche_net_opt_valid(options=options, process=process)
    # GNN options
    options = gnn_opt_valid(options=options, process=process)
    # niche trajectory
    options = nt_opt_valid(options=options, process=process)

    info('------------------ RUN params memo ------------------ ')
    write_io_options_memo(options=options, io_options=io_options)
    write_niche_net_constr_memo(options=options)
    write_train_options_memo(options=options)
    write_GCN_options_memo(options=options)
    write_GP_options_memo(options=options)
    write_NT_options_memo(options=options)
    info('--------------- RUN params memo end ----------------- ')

    return options


def run_ontrac(options: Values) -> None:
    """
    Run ONTraC
    :param options: options
    :return: None
    """

    # ----- options validation -----
    options = options_valid(options=options)

    # ----- Niche Network Construct -----
    niche_network_construct(options=options)

    # ----- GNN -----
    gnn(options=options)

    # ----- NT score -----
    niche_trajectory_construct(options=options)



from pathlib import Path
from typing import List, Optional, Tuple, Union

import matplotlib as mpl
import networkx as nx
import numpy as np
import pandas as pd
from matplotlib.axes import Axes

mpl.rcParams['pdf.fonttype'] = 42
mpl.rcParams['ps.fonttype'] = 42
mpl.rcParams['font.family'] = 'Arial'

import matplotlib.gridspec as gridspec
import matplotlib.pyplot as plt
import seaborn as sns
from matplotlib.cm import ScalarMappable
from matplotlib.colors import LinearSegmentedColormap, Normalize

from ..log import warning
from .data import AnaData
from .utils import gini, saptial_figsize


def plot_niche_cluster_connectivity(
        niche_cluster_connectivity: np.ndarray,
        niche_cluster_score: Optional[np.ndarray] = None,
        niche_cluster_size: Optional[np.ndarray] = None,
        reverse: bool = False,
        output_file_path: Optional[Union[str, Path]] = None) -> Optional[Tuple[plt.Figure, List[plt.Axes]]]:
    """
    Plot niche cluster connectivity.
    :param niche_cluster_connectivity: np.ndarray, the connectivity matrix.
    :param niche_cluster_score: Optional[np.ndarray], the score of each niche cluster.
    :param niche_cluster_size: Optional[np.ndarray], the size of each niche cluster.
    :param reverse: bool, whether to reverse the color.
    :param output_file_path: Optional[Union[str, Path]], the output file path.
    :return: None or Tuple[plt.Figure, plt.Axes].
    """

    G = nx.Graph(niche_cluster_connectivity / niche_cluster_connectivity.max())

    # position
    pos = nx.spring_layout(G, seed=42)
    # edges
    edges = G.edges()
    weights = [G[u][v]['weight'] for u, v in edges]
    # node color
    if niche_cluster_score is not None:
        norm = Normalize(vmin=0, vmax=1)
        sm = ScalarMappable(cmap=plt.cm.rainbow, norm=norm)  # type: ignore
        nc_scores = 1 - niche_cluster_score if reverse else niche_cluster_score
        niche_cluster_colors = [sm.to_rgba(nc_scores[n]) for n in G.nodes]
    else:
        niche_cluster_colors = ["#1f78b4"] * niche_cluster_connectivity.shape[0]
    # node size
    if niche_cluster_size is None:
        node_size = 500
    else:
        # rescale the node with size 0 to 200, and the maximum size is 1000
        node_size = (niche_cluster_size / niche_cluster_size.max() * 800 + 200).to_list()

    # Create a figure
    ## figwidth
    figwidths = [6]
    if niche_cluster_size is not None:
        figwidths.append(1.5)
    if niche_cluster_score is not None:
        figwidths.append(.5)
    figwidths.append(.5)
    fig = plt.figure(figsize=(sum(figwidths), 6))

    # Create a gridspec with 1 row and 2 columns, with widths of A and B
    axes = []
    gs = gridspec.GridSpec(nrows=1, ncols=len(figwidths), width_ratios=figwidths)
    ax_index = 0
    graph_ax = fig.add_subplot(gs[ax_index])
    axes.append(graph_ax)
    if niche_cluster_size is not None:
        ax_index += 1
        node_size_ax = fig.add_subplot(gs[ax_index])
        axes.append(node_size_ax)
    if niche_cluster_score is not None:
        ax_index += 1
        node_colorbar_ax = fig.add_subplot(gs[ax_index])
        axes.append(node_colorbar_ax)
    ax_index += 1
    edge_colorbar_ax = fig.add_subplot(gs[ax_index])
    axes.append(edge_colorbar_ax)

    # Draw the graph
    nx.draw_networkx_nodes(G=G, pos=pos, node_color=niche_cluster_colors, node_size=node_size, ax=graph_ax)
    nx.draw_networkx_edges(
        G=G,
        pos=pos,
        edge_color=weights,
        alpha=weights,
        width=3.0,
        edge_cmap=plt.cm.Reds,  # type: ignore
        ax=graph_ax,
        node_size=node_size)
    nx.draw_networkx_labels(G, pos, ax=graph_ax)
    graph_ax.axis('off')

    # Draw legend for node size
    if niche_cluster_size is not None:
        ## prepare
        node_size_ax.axis('off')  # hide the
        node_size_legend_num = 5  # TODO: make it become a parameter
        sizes = np.linspace(start=200, stop=1000, num=node_size_legend_num)
        max_niche_cluster_size = niche_cluster_size.values.max()
        magnitude = 10**int(np.floor(np.log10(abs(max_niche_cluster_size))))
        max_label = round(max_niche_cluster_size / magnitude) * magnitude
        labels = [f'{int(x):d}' for x in np.linspace(0, max_label, node_size_legend_num)]
        ## draw legend
        handles = [
            plt.Line2D([0], [0], marker='o', color='w', markerfacecolor='gray', markersize=np.sqrt(size), label=label)
            for size, label in zip(sizes, labels)
        ]
        legend = node_size_ax.legend(handles=handles,
                                     title='Niche cluster size',
                                     loc='center',
                                     frameon=False,
                                     bbox_to_anchor=(0.6, 0.6))
        ## Rotate the title and adjust the position
        title = legend.get_title()
        title.set_rotation(90)
        title.set_verticalalignment('center')  # Align title vertically
        title.set_horizontalalignment('right')  # Align title horizontally
        title.set_position((-70, -60))  # Shift the title slightly to the right

    # Draw the colorbar for nodes
    if niche_cluster_score is not None:
        gradient = np.linspace(0, 1, 1000).reshape(-1, 1)
        node_colorbar_ax.imshow(gradient, aspect='auto', cmap=plt.cm.rainbow)
        node_colorbar_ax.set_xticks([])
        node_colorbar_ax.set_yticks(np.linspace(0, 1000, 5))
        node_colorbar_ax.set_yticklabels(f'{x:.2f}' for x in np.linspace(0, 1, 5))
        node_colorbar_ax.set_ylabel('NT score (nodes)')

    # Draw the colorbar for edges
    colors = [(1, 1, 1, 0), (1, 0, 0, 1)]
    custom_cmap = LinearSegmentedColormap.from_list('custom_cmap', colors)
    gradient = np.linspace(1, 0, 1000).reshape(-1, 1)
    edge_colorbar_ax.imshow(gradient, aspect='auto', cmap=custom_cmap)
    edge_colorbar_ax.set_xticks([])
    edge_colorbar_ax.set_yticks(np.linspace(1000, 0, 5))
    edge_colorbar_ax.set_yticklabels(f'{x:.2f}' for x in np.linspace(0, niche_cluster_connectivity.max(), 5))
    edge_colorbar_ax.set_ylabel('Connectivity (edges)')

    fig.suptitle('Niche cluster connectivity')
    fig.tight_layout()

    if output_file_path is not None:
        fig.savefig(output_file_path)
        plt.close(fig)
        return None
    else:
        return fig, axes


def plot_niche_cluster_connectivity_from_anadata(ana_data: AnaData) -> Optional[Tuple[plt.Figure, List[plt.Axes]]]:
    """
    Plot niche cluster connectivity.
    :param ana_data: AnaData, the data for analysis.
    :return: None or Tuple[plt.Figure, plt.Axes].
    """
    try:
        if ana_data.niche_cluster_score is None or ana_data.niche_cluster_connectivity is None:
            warning("No niche cluster score or connectivity data found.")
            return None
    except FileNotFoundError as e:
        warning(str(e))
        return None

    if not hasattr(ana_data, 'cell_level_niche_cluster_assign') or ana_data.cell_level_niche_cluster_assign is None:
        niche_cluster_size = np.ones(ana_data.cell_level_niche_cluster_assign.shape[1])
    else:
        niche_cluster_size = ana_data.cell_level_niche_cluster_assign.sum(axis=0)

    output_file_path = f'{ana_data.options.output}/niche_cluster_connectivity.pdf' if ana_data.options.output is not None else None

    return plot_niche_cluster_connectivity(
        niche_cluster_connectivity=ana_data.niche_cluster_connectivity,
        niche_cluster_score=ana_data.niche_cluster_score,
        niche_cluster_size=niche_cluster_size,  # type: ignore
        reverse=ana_data.options.reverse,
        output_file_path=output_file_path)


def plot_niche_cluster_connectivity_bysample_from_anadata(ana_data: AnaData) -> None:
    """
    Plot niche cluster connectivity by sample.
    :param ana_data: AnaData, the data for analysis.
    :return: None.
    """

    for sample in ana_data.meta_data_df['Sample'].unique():

        niche_cluster_connectivity = np.loadtxt(f'{ana_data.options.GNN_dir}/{sample}_out_adj.csv.gz', delimiter=',')

        cell_level_niche_cluster_assign = np.loadtxt(f'{ana_data.options.GNN_dir}/{sample}_s.csv.gz', delimiter=',')
        niche_cluster_size = cell_level_niche_cluster_assign.sum(axis=0)

        output_file_path = f'{ana_data.options.output}/{sample}_cluster_connectivity.pdf'

        plot_niche_cluster_connectivity(niche_cluster_connectivity=niche_cluster_connectivity,
                                        niche_cluster_score=ana_data.niche_cluster_score,
                                        niche_cluster_size=niche_cluster_size,
                                        reverse=ana_data.options.reverse,
                                        output_file_path=output_file_path)


def plot_cluster_proportion(
        niche_cluster_loading: pd.DataFrame,
        niche_cluster_colors: List,
        output_file_path: Optional[Union[str, Path]] = None) -> Optional[Tuple[plt.Figure, plt.Axes]]:
    """
    Plot the proportion of each cluster.
    :param niche_cluster_loading: pd.DataFrame, the loading of each niche cluster.
    :param niche_cluster_colors: List, the color of each niche cluster.
    :param output_file_path: Optional[Union[str, Path]], the output file path.
    :return: None or Tuple[plt.Figure, plt.Axes].
    """

    fig, ax = plt.subplots(figsize=(6, 6))
    ax.pie(niche_cluster_loading,
           labels=[f'Niche cluster {i}' for i in range(niche_cluster_loading.shape[0])],
           colors=niche_cluster_colors,
           autopct='%1.1f%%',
           pctdistance=1.25,
           labeldistance=.6)
    ax.set_title(f'Niche proportions for each niche cluster')
    fig.tight_layout()
    if output_file_path is not None:
        fig.savefig(f'{output_file_path}/niche_cluster_proportion.pdf')
        plt.close(fig)
        return None
    else:
        return fig, ax


def plot_cluster_proportion_from_anadata(ana_data: AnaData) -> Optional[Tuple[plt.Figure, plt.Axes]]:
    """
    Plot the proportion of each cluster.
    :param ana_data: AnaData, the data for analysis.
    :return: None or Tuple[plt.Figure, plt.Axes].
    """
    try:
        if ana_data.niche_cluster_score is None or ana_data.niche_level_niche_cluster_assign is None:
            warning("No niche cluster score or connectivity data found.")
            return None
    except FileNotFoundError as e:
        warning(str(e))
        return None

    # colors
    norm = Normalize(vmin=0, vmax=1)
    sm = ScalarMappable(cmap=plt.cm.rainbow, norm=norm)  # type: ignore
    nc_scores = 1 - ana_data.niche_cluster_score if ana_data.options.reverse else ana_data.niche_cluster_score
    niche_cluster_colors = [sm.to_rgba(nc_scores[n]) for n in np.arange(ana_data.niche_cluster_score.shape[0])]

    # loadings
    niche_cluster_loading = ana_data.niche_level_niche_cluster_assign.sum(axis=0)

    return plot_cluster_proportion(niche_cluster_loading=niche_cluster_loading,
                                   niche_cluster_colors=niche_cluster_colors,
                                   output_file_path=ana_data.options.output)


def plot_niche_cluster_loadings_dataset(
    cell_level_niche_cluster_assign: pd.DataFrame,
    meta_data_df: pd.DataFrame,
    nc_scores: np.ndarray,
    output_file_path: Optional[Union[str,
                                     Path]] = None) -> Optional[Tuple[plt.Figure, Union[plt.Axes, List[plt.Axes]]]]:
    """
    Plot niche cluster loadings for each cell.
    :param cell_level_niche_cluster_assign: pd.DataFrame, the niche cluster assign data.
    :param meta_data_df: pd.DataFrame, the meta data.
    :param nc_scores: np.ndarray, the score of each niche cluster.
    :param output_file_path: Optional[Union[str, Path]], the output file path.
    :return: None or Tuple[plt.Figure, plt.Axes].
    """

    samples = meta_data_df['Sample'].unique()
    n_sample = len(samples)
    n_niche_cluster = cell_level_niche_cluster_assign.shape[1]

    fig, axes = plt.subplots(n_sample, n_niche_cluster, figsize=(3.3 * n_niche_cluster, 3 * n_sample))
    for i, sample in enumerate(samples):
        sample_df = cell_level_niche_cluster_assign.loc[meta_data_df[meta_data_df['Sample'] == sample].index]
        sample_df = sample_df.join(meta_data_df[['x', 'y']])
        for j, c_index in enumerate(nc_scores.argsort()):
            ax = axes[i, j] if n_sample > 1 else axes[j]
            scatter = ax.scatter(sample_df['x'],
                                 sample_df['y'],
                                 c=sample_df[f'NicheCluster_{c_index}'],
                                 cmap='Reds',
                                 vmin=0,
                                 vmax=1,
                                 s=4)
            ax.set_title(f'{sample}: niche cluster {c_index}')
            plt.colorbar(scatter)
    fig.tight_layout()
    if output_file_path is not None:
        fig.savefig(f'{output_file_path}/niche_cluster_loadings.pdf')
        plt.close(fig)
        return None
    else:
        return fig, axes


def plot_niche_cluster_loadings_dataset_from_anadata(
        ana_data: AnaData) -> Optional[Tuple[plt.Figure, Union[plt.Axes, List[plt.Axes]]]]:
    """
    Plot niche cluster loadings for each cell.
    :param ana_data: AnaData, the data for analysis.
    :return: None or Tuple[plt.Figure, Union[plt.Axes, List[plt.Axes]]].
    """

    try:
        if ana_data.niche_cluster_score is None or ana_data.cell_type_composition is None or ana_data.cell_level_niche_cluster_assign is None:
            warning("No niche cluster score, cell type composition or cluster assign data found.")
            return None
    except FileNotFoundError as e:
        warning(str(e))
        return None

    nc_scores = 1 - ana_data.niche_cluster_score if ana_data.options.reverse else ana_data.niche_cluster_score

    return plot_niche_cluster_loadings_dataset(cell_level_niche_cluster_assign=ana_data.cell_level_niche_cluster_assign,
                                               meta_data_df=ana_data.meta_data_df,
                                               nc_scores=nc_scores,
                                               output_file_path=ana_data.options.output)


def plot_niche_cluster_loadings_sample(
        cell_level_niche_cluster_assign: pd.DataFrame,
        meta_data_df: pd.DataFrame,
        nc_scores: np.ndarray,
        spatial_scaling_factor: float = 1.0,
        output_file_path: Optional[Union[str, Path]] = None) -> Optional[List[Tuple[plt.Figure, plt.Axes]]]:
    """
    Plot niche cluster loadings for each cell.
    :param cell_level_niche_cluster_assign: pd.DataFrame, the niche cluster assign data.
    :param meta_data_df: pd.DataFrame, the meta data.
    :param nc_scores: np.ndarray, the score of each niche cluster.
    :param spatial_scaling_factor: float, the scale factor control the size of spatial-based plots.
    :param output_file_path: Optional[Union[str, Path]], the output file path.
    :return: None or List[Tuple[plt.Figure, plt.Axes]].
    """

    samples = meta_data_df['Sample'].unique()

    output = []
    for sample in samples:
        sample_df = cell_level_niche_cluster_assign.loc[meta_data_df[meta_data_df['Sample'] == sample].index]
        sample_df = sample_df.join(meta_data_df[['x', 'y']])
        fig_width, fig_height = saptial_figsize(sample_df, scaling_factor=spatial_scaling_factor)
        fig, axes = plt.subplots(1, nc_scores.shape[0], figsize=(fig_width * nc_scores.shape[0], fig_height))
        for j, c_index in enumerate(nc_scores.argsort()):
            ax = axes[j]  #  there should more than one niche cluster
            scatter = ax.scatter(sample_df['x'],
                                 sample_df['y'],
                                 c=sample_df[f'NicheCluster_{c_index}'],
                                 cmap='Reds',
                                 vmin=0,
                                 vmax=1,
                                 s=4)
            ax.set_title(f'{sample}: niche cluster {c_index}')
            plt.colorbar(scatter)
        fig.tight_layout()
        output.append((fig, axes))
        if output_file_path is not None:
            fig.savefig(f'{output_file_path}/niche_cluster_loadings_{sample}.pdf')
            plt.close(fig)
    return output if len(output) > 0 else None


def plot_niche_cluster_loadings_sample_from_anadata(ana_data: AnaData) -> Optional[List[Tuple[plt.Figure, plt.Axes]]]:
    """
    Plot niche cluster loadings for each cell.
    :param ana_data: AnaData, the data for analysis.
    :return: None or List[Tuple[plt.Figure, plt.Axes]].
    """

    try:
        if ana_data.niche_cluster_score is None or ana_data.cell_type_composition is None or ana_data.cell_level_niche_cluster_assign is None:
            warning("No niche cluster score, cell type composition or cluster assign data found.")
            return None
    except FileNotFoundError as e:
        warning(str(e))
        return None

    nc_scores = 1 - ana_data.niche_cluster_score if ana_data.options.reverse else ana_data.niche_cluster_score

    return plot_niche_cluster_loadings_sample(cell_level_niche_cluster_assign=ana_data.cell_level_niche_cluster_assign,
                                              meta_data_df=ana_data.meta_data_df,
                                              nc_scores=nc_scores,
                                              spatial_scaling_factor=ana_data.options.scale_factor,
                                              output_file_path=ana_data.options.output)


def plot_niche_cluster_loadings(
    ana_data: AnaData
) -> Optional[Union[List[Tuple[plt.Figure, plt.Axes]], Tuple[plt.Figure, Union[plt.Axes, List[plt.Axes]]]]]:
    """
    Plot niche cluster loadings for each cell.
    :param ana_data: AnaData, the data for analysis.
    :return: None or Union[List[Tuple[plt.Figure, plt.Axes]], Tuple[plt.Figure, Union[plt.Axes, List[plt.Axes]]]].
    """
    if hasattr(ana_data.options, 'sample') and ana_data.options.sample:
        return plot_niche_cluster_loadings_sample_from_anadata(ana_data=ana_data)
    else:
        return plot_niche_cluster_loadings_dataset_from_anadata(ana_data=ana_data)


def plot_max_niche_cluster_dataset(
        cell_level_max_niche_cluster: pd.DataFrame,
        meta_data_df: pd.DataFrame,
        nc_scores: np.ndarray,
        output_file_path: Optional[Union[str, Path]] = None) -> Optional[Tuple[plt.Figure, plt.Axes]]:
    """
    Plot the maximum niche cluster for each cell.
    :param cell_level_max_niche_cluster: pd.DataFrame, the maximum niche cluster data.
    :param meta_data_df: pd.DataFrame, the meta data.
    :param nc_scores: np.ndarray, the score of each niche cluster.
    :param output_file_path: Optional[Union[str, Path]], the output file path.
    :return: None or Tuple[plt.Figure, plt.Axes].
    """

    samples = meta_data_df['Sample'].unique()
    n_sample = len(samples)
    n_niche_cluster = len(nc_scores)

    # colors
    norm = Normalize(vmin=0, vmax=1)
    sm = ScalarMappable(cmap=plt.cm.rainbow, norm=norm)  # type: ignore
    niche_cluster_colors = [sm.to_rgba(nc_scores[n]) for n in np.arange(n_niche_cluster)]
    palette = {f'niche cluster {i}': niche_cluster_colors[i] for i in range(n_niche_cluster)}

    fig, axes = plt.subplots(1, n_sample, figsize=(5 * n_sample, 3))
    for i, sample in enumerate(samples):
        ax: Axes = axes[i] if n_sample > 1 else axes  # type: ignore
        sample_df = cell_level_max_niche_cluster.loc[meta_data_df[meta_data_df['Sample'] == sample].index]
        sample_df = sample_df.join(meta_data_df[['x', 'y']])
        sample_df['Niche_Cluster'] = 'niche cluster ' + sample_df['Niche_Cluster'].astype(str)
        sns.scatterplot(data=sample_df,
                        x='x',
                        y='y',
                        hue='Niche_Cluster',
                        hue_order=[f'niche cluster {j}' for j in nc_scores.argsort()],
                        palette=palette,
                        s=10,
                        ax=ax)
        ax.set_title(f'{sample}')
        ax.legend(loc='upper left', bbox_to_anchor=(1, 1))
    fig.tight_layout()
    if output_file_path is not None:
        fig.savefig(f'{output_file_path}/max_niche_cluster.pdf')
        plt.close(fig)
        return None
    else:
        return fig, axes


def plot_max_niche_cluster_dataset_from_anadata(ana_data: AnaData) -> Optional[Tuple[plt.Figure, plt.Axes]]:
    """
    Plot the maximum niche cluster for each cell.
    :param ana_data: AnaData, the data for analysis.
    :return: None or Tuple[plt.Figure, plt.Axes].
    """

    try:
        if ana_data.niche_cluster_score is None or ana_data.cell_type_composition is None or ana_data.cell_level_max_niche_cluster is None:
            warning("No niche cluster score, cell type composition or max niche cluster data found.")
            return None
    except FileNotFoundError as e:
        warning(str(e))
        return None

    nc_scores = 1 - ana_data.niche_cluster_score if ana_data.options.reverse else ana_data.niche_cluster_score

    return plot_max_niche_cluster_dataset(cell_level_max_niche_cluster=ana_data.cell_level_max_niche_cluster,
                                          meta_data_df=ana_data.meta_data_df,
                                          nc_scores=nc_scores,
                                          output_file_path=ana_data.options.output)


def plot_max_niche_cluster_sample(
        cell_level_max_niche_cluster: pd.DataFrame,
        meta_data_df: pd.DataFrame,
        nc_scores: np.ndarray,
        output_file_path: Optional[Union[str, Path]] = None) -> Optional[List[Tuple[plt.Figure, plt.Axes]]]:
    """
    Plot the maximum niche cluster for each cell.
    :param cell_level_max_niche_cluster: pd.DataFrame, the maximum niche cluster data.
    :param meta_data_df: pd.DataFrame, the meta data.
    :param nc_scores: np.ndarray, the score of each niche cluster.
    :param output_file_path: Optional[Union[str, Path]], the output file path.
    :return: None or List[Tuple[plt.Figure, plt.Axes]].
    """

    samples = meta_data_df['Sample'].unique()
    n_niche_cluster = len(nc_scores)

    # colors
    norm = Normalize(vmin=0, vmax=1)
    sm = ScalarMappable(cmap=plt.cm.rainbow, norm=norm)  # type: ignore
    niche_cluster_colors = [sm.to_rgba(nc_scores[n]) for n in np.arange(n_niche_cluster)]
    palette = {f'niche cluster {i}': niche_cluster_colors[i] for i in range(n_niche_cluster)}

    output = []
    for sample in samples:
        sample_df = cell_level_max_niche_cluster.loc[meta_data_df[meta_data_df['Sample'] == sample].index]
        sample_df = sample_df.join(meta_data_df[['x', 'y']])
        sample_df['Niche_Cluster'] = 'niche cluster ' + sample_df['Niche_Cluster'].astype(str)
        fig_width, fig_height = saptial_figsize(sample_df, scaling_factor=1)
        fig, ax = plt.subplots(1, 1, figsize=(fig_width, fig_height))
        sns.scatterplot(data=sample_df,
                        x='x',
                        y='y',
                        hue='Niche_Cluster',
                        hue_order=[f'niche cluster {j}' for j in nc_scores.argsort()],
                        palette=palette,
                        s=10,
                        ax=ax)
        ax.set_title(f'{sample}')
        ax.legend(loc='upper left', bbox_to_anchor=(1, 1))
        fig.tight_layout()
        if output_file_path is not None:
            fig.savefig(f'{output_file_path}/max_niche_cluster_{sample}.pdf')
            plt.close(fig)
        else:
            output.append((fig, ax))
    return output if len(output) > 0 else None


def plot_max_niche_cluster_sample_from_anadata(ana_data: AnaData) -> Optional[List[Tuple[plt.Figure, plt.Axes]]]:
    """
    Plot the maximum niche cluster for each cell.
    :param ana_data: AnaData, the data for analysis.
    :return: None or List[Tuple[plt.Figure, plt.Axes]].
    """

    try:
        if ana_data.niche_cluster_score is None or ana_data.cell_type_composition is None or ana_data.cell_level_max_niche_cluster is None:
            warning("No niche cluster score, cell type composition or max niche cluster data found.")
            return None
    except FileNotFoundError as e:
        warning(str(e))
        return None

    nc_scores = 1 - ana_data.niche_cluster_score if ana_data.options.reverse else ana_data.niche_cluster_score

    return plot_max_niche_cluster_sample(cell_level_max_niche_cluster=ana_data.cell_level_max_niche_cluster,
                                         meta_data_df=ana_data.meta_data_df,
                                         nc_scores=nc_scores,
                                         output_file_path=ana_data.options.output)


def plot_max_niche_cluster(
        ana_data: AnaData) -> Optional[Union[List[Tuple[plt.Figure, plt.Axes]], Tuple[plt.Figure, plt.Axes]]]:
    """
    Plot the maximum niche cluster for each cell.
    :param ana_data: AnaData, the data for analysis.
    :return: None or Tuple[plt.Figure, plt.Axes].
    """
    if hasattr(ana_data.options, 'sample') and ana_data.options.sample:
        return plot_max_niche_cluster_sample_from_anadata(ana_data=ana_data)
    else:
        return plot_max_niche_cluster_dataset_from_anadata(ana_data=ana_data)


def plot_niche_cluster_gini(
        intra_cluster_gini_df: pd.DataFrame,
        output_file_path: Optional[Union[str, Path]] = None) -> Optional[Tuple[plt.Figure, plt.Axes]]:
    """
    Plot the Gini coefficient of each niche cluster.
    :param intra_cluster_gini_df: pd.DataFrame, the Gini coefficient of each niche cluster.
    :param output_file_path: Optional[Union[str, Path]], the output file path.
    :return: None or Tuple[plt.Figure, plt.Axes].
    """

    fig, ax = plt.subplots(figsize=(6, 4))
    sns.barplot(data=intra_cluster_gini_df, x='cluster', y='gini', ax=ax)
    ax.set_xlabel('Niche Cluster')
    ax.set_ylabel('Gini coefficient across each cell')
    ax.set_xticklabels(ax.get_xticklabels(), rotation=45, horizontalalignment='right')
    fig.tight_layout()
    if output_file_path is not None:
        fig.savefig(f'{output_file_path}/niche_cluster_gini.pdf')
        plt.close(fig)
        return None
    else:
        return fig, ax


def plot_niche_cluster_gini_from_anadata(ana_data: AnaData) -> Optional[Tuple[plt.Figure, plt.Axes]]:
    """
    Plot the Gini coefficient of each niche cluster.
    :param ana_data: AnaData, the data for analysis.
    :return: None or Tuple[plt.Figure, plt.Axes].
    """
    try:
        if ana_data.cell_level_niche_cluster_assign is None:
            warning("No niche cluster assign data found.")
            return None
    except FileNotFoundError as e:
        warning(str(e))
        return None
    intra_cluster_gini = ana_data.cell_level_niche_cluster_assign.apply(gini, axis=0).values
    intra_cluster_gini_df = pd.DataFrame(data={
        'gini': intra_cluster_gini,
        'cluster': ana_data.cell_level_niche_cluster_assign.columns
    })
    fig, ax = plt.subplots(figsize=(6, 4))
    sns.barplot(data=intra_cluster_gini_df, x='cluster', y='gini', ax=ax)
    ax.set_xlabel('Niche Cluster')
    ax.set_ylabel('Gini coefficient across each cell')
    ax.set_xticklabels(ax.get_xticklabels(), rotation=45, horizontalalignment='right')
    fig.tight_layout()
    if ana_data.options.output is not None:
        fig.savefig(f'{ana_data.options.output}/niche_cluster_gini.pdf')
        plt.close(fig)
        return None
    else:
        return fig, ax


def niche_cluster_visualization(ana_data: AnaData) -> None:
    """
    All spatial visualization will include here.
    :param ana_data: AnaData, the data for analysis.
    :return: None.
    """

    # 1. plot niche cluster connectivity
    plot_niche_cluster_connectivity_from_anadata(ana_data=ana_data)

    if hasattr(ana_data.options, 'sample') and ana_data.options.sample:
        plot_niche_cluster_connectivity_bysample_from_anadata(ana_data=ana_data)

    # 2. share of each cluster
    plot_cluster_proportion_from_anadata(ana_data=ana_data)

    # 3. niche cluster loadings for each cell
    if not hasattr(ana_data.options,
                   'suppress_niche_cluster_loadings') or not ana_data.options.suppress_niche_cluster_loadings:
        plot_niche_cluster_loadings(ana_data=ana_data)

    # 4. maximum niche cluster for each cell
    plot_max_niche_cluster(ana_data=ana_data)

    # 5. gini coefficient of each niche cluster
    plot_niche_cluster_gini_from_anadata(ana_data=ana_data)



import torch
from torch import Tensor
from torch.nn import Parameter

from torch_geometric.nn.dense.linear import Linear
from torch_geometric.nn.inits import zeros
from torch_geometric.typing import OptTensor

from ..log import *


class NormDenseGCNConv(torch.nn.Module):
    r"""See :class:`torch_geometric.nn.conv.GCNConv`."""

    def __init__(
        self,
        in_channels: int,
        out_channels: int,
        improved: bool = False,
        bias: bool = True,
    ):
        super().__init__()

        self.in_channels = in_channels
        self.out_channels = out_channels
        self.improved = improved

        self.lin = Linear(in_channels, out_channels, bias=False, weight_initializer='glorot')

        if bias:
            self.bias = Parameter(torch.Tensor(out_channels))
        else:
            self.register_parameter('bias', None)

        self.reset_parameters()

    def reset_parameters(self):
        r"""Resets all learnable parameters of the module."""
        self.lin.reset_parameters()
        zeros(self.bias)

    def forward(self, x: Tensor, adj: Tensor, mask: OptTensor = None) -> Tensor:
        r"""
        :math: `\mathbf{\hat{L}}X\mathbf{\Theta}`, with
               `\mathbf{\hat{L}} = \mathbf{D}^{-1/2}\mathbf{A}\mathbf{D}^{-1/2} + I`

        Args:
            x (torch.Tensor): Node feature tensor
                :math:`\mathbf{X} \in \mathbb{R}^{B \times N \times F}`, with
                batch-size :math:`B`, (maximum) number of nodes :math:`N` for
                each graph, and feature dimension :math:`F`.
            adj (torch.Tensor): Adjacency tensor
                :math:`\mathbf{A} \in \mathbb{R}^{B \times N \times N}`.
                The adjacency tensor is broadcastable in the batch dimension,
                resulting in a shared adjacency matrix for the complete batch.
            mask (torch.Tensor, optional): Mask matrix
                :math:`\mathbf{M} \in {\{ 0, 1 \}}^{B \times N}` indicating
                the valid nodes for each graph. (default: :obj:`None`)
        """

        # inputs shape check
        x = x.unsqueeze(0) if x.dim() == 2 else x
        adj = adj.unsqueeze(0) if adj.dim() == 2 else adj
        B, N, _ = adj.size()

        # X\mathbf{\Theta}
        out = self.lin(x)

        # calculate \mathbf{\hat{L}} = \mathbf{D}^{-1/2}\mathbf{A}\mathbf{D}^{-1/2} + I
        deg_inv_sqrt = adj.sum(dim=-1).clamp(min=1).pow(-0.5)
        adj = deg_inv_sqrt.unsqueeze(-1) * adj * deg_inv_sqrt.unsqueeze(-2)
        eye_matrix = torch.eye(N).to(adj.device).unsqueeze(0).repeat(B, 1, 1)
        laplacian = adj + eye_matrix if not self.improved else adj + 2 * eye_matrix

        # \mathbf{\hat{L}}X\mathbf{\Theta}
        out = torch.matmul(laplacian, out)

        if self.bias is not None:
            out = out + self.bias

        if mask is not None:
            out = out * mask.view(B, N, 1).to(x.dtype)

        return out

    def __repr__(self) -> str:
        return (f'{self.__class__.__name__}({self.in_channels}, '
                f'{self.out_channels})')



from typing import Dict, List

IO_OPTIONS: Dict[str, List[str]] = {
    'ONTraC': ['NN_dir', 'GNN_dir', 'NT_dir', 'input'],
    'ONTraC_NN': ['NN_dir', 'input'],  # niche network
    'ONTraC_GNN': ['NN_dir', 'GNN_dir'],  # graph neural network
    'ONTraC_NT': ['NN_dir', 'GNN_dir', 'NT_dir'],  # niche trajectory
    'ONTraC_GT': ['NN_dir', 'GNN_dir', 'NT_dir'],  # GNN & niche trajectory
    'ONTraC_analysis': ['NN_dir', 'GNN_dir', 'NT_dir', 'input', 'output', 'log'],  # analysis
}



from optparse import OptionParser, Values

from ..log import *
from ..version import __version__
from ._IO import *
from ._train import *


# ------------------------------------
# Functions
# ------------------------------------
def add_NT_options_group(optparser: OptionParser) -> None:
    """
    Add niche trajectory options group to optparser.
    :param optparser: OptionParser object.
    :return: OptionGroup object.
    """

    group_NT = OptionGroup(optparser, "Options for niche trajectory")
    optparser.add_option_group(group_NT)
    group_NT.add_option(
        '--trajectory-construct',
        dest='trajectory_construct',
        default='BF',
        choices=['BF', 'TSP'],
        help="Method to construct the niche trajectory. Default is 'BF' (brute-force). A faster alternative is 'TSP'.")


def validate_NT_options(optparser: OptionParser, options: Values) -> None:
    """
    Validate niche trajectory options.
    Placehold and do nothing.

    :param optparser: OptionParser object.
    :param options: Options object.
    :return: None.
    """

    pass


def write_NT_options_memo(options: Values) -> None:
    """
    Write niche trajectory options memo.
    :param options: Values, options.
    :return: None.
    """

    info('---------------- Niche trajectory options ----------------')
    info(f'Niche trajectory construction method: {options.trajectory_construct}')



from typing import Union, Tuple

import numpy as np
import pandas as pd


def saptial_figsize(sample_df, scaling_factor: Union[int, float]=1) -> Tuple[int, int]:
    """
    Calculate the figure size for spatial-based plot according to the points and the span of x and y.
    :param sample_df: pd.DataFrame, the sample data.
    :param scale_factor: float, the scale factor control the size of spatial-based plots. The larger the scale factor,
    the larger the figure size.
    :return: tuple[int, int], the figure size.
    """

    n_points = sample_df[['x', 'y']].dropna().shape[0]
    # debug(f'n_points: {n_points}')

    x_span = sample_df['x'].dropna().max() - sample_df['x'].dropna().min()
    y_span = sample_df['y'].dropna().max() - sample_df['y'].dropna().min()
    # debug(f'x_span: {x_span}')
    # debug(f'y_span: {y_span}')

    points_density = n_points / x_span / y_span * 10_000

    # debug(f'points density: {points_density}')

    fig_width = x_span / 2_000 * scaling_factor * np.sqrt(points_density) + .5  # Adding 2 for colorbar space
    fig_height = y_span / 2_000 * scaling_factor * np.sqrt(points_density) + .2  # Adding 1.5 for title space

    # debug(f'scale_factor: {scale_factor}')
    # debug(f'fig_width: {fig_width}')
    # debug(f'fig_height: {fig_height}')

    return fig_width, fig_height


def gini(array: Union[np.ndarray, pd.Series]) -> float:
    """Calculate the Gini coefficient of a numpy array.
    :param array: np.ndarray or pd.Series, the array for calculating Gini coefficient.
    :return: float, the Gini coefficient.
    """
    #
    # from:
    # http://www.statsdirect.com/help/default.htm#nonparametric_methods/gini.htm
    # All values are treated equally, arrays must be 1d:
    if isinstance(array, pd.Series):
        array = np.array(array)
    array = array.flatten()  # type: ignore
    if np.amin(array) < 0:
        # Values cannot be negative:
        array -= np.amin(array)  # type: ignore
    # Values cannot be 0:
    array += 0.0000001
    # Values must be sorted:
    array = np.sort(array)  # type: ignore
    # Number of array elements:
    n = array.shape[0]  # type: ignore
    # Index per array element:
    index = np.arange(1, n + 1)  # type: ignore
    # Gini coefficient:
    return ((np.sum((2 * index - n - 1) * array)) / (n * np.sum(array)))  # type: ignore



from optparse import Values

from ..analysis.cell_type import cell_type_visualization
from ..analysis.data import AnaData
from ..analysis.niche_cluster import niche_cluster_visualization
from ..analysis.spatial import spatial_visualization
from ..analysis.train_loss import train_loss_visualiztion
from ..log import *
from ..optparser import opt_analysis_validate, prepare_analysis_optparser
from ..run.processes import load_parameters
from ..utils import *
from ..version import __version__


# ------------------------------------
# Functions
# ------------------------------------
def analysis_pipeline(options: Values) -> None:
    """
    ONTraC analysis pipeline
    """

    # 0. load data class
    ana_data = AnaData(options)

    # part 1: train loss
    train_loss_visualiztion(ana_data=ana_data)

    # part 2: spatial based output
    spatial_visualization(ana_data=ana_data)

    # part 3: niche cluster
    niche_cluster_visualization(ana_data=ana_data)

    # part 4: cell type based output
    cell_type_visualization(ana_data=ana_data)


def main():
    """
    The main function
    """

    # write version information
    write_version_info()

    options = load_parameters(opt_validate_func=opt_analysis_validate,
                              prepare_optparser_func=prepare_analysis_optparser)

    analysis_pipeline(options)


if __name__ == '__main__':
    main()



import os
import re
import sys
from optparse import OptionGroup, OptionParser, Values
from typing import List, Optional

from ..log import *


def add_IO_options_group(optparser: OptionParser, io_options: Optional[List[str]]) -> None:
    """
    Add I/O options group to optparser.
    :param optparser: OptionParser object.
    :param io_options: List of I/O options.
    :return: OptionGroup object.
    """
    if io_options is None:
        return
    # I/O options group
    group_io = OptionGroup(optparser, "IO")

    # directories
    if 'NN_dir' in io_options:
        group_io.add_option('--NN-dir', dest='NN_dir', type='string', help='Directory for niche network outputs.')
    if 'GNN_dir' in io_options:
        group_io.add_option('--GNN-dir', dest='GNN_dir', type='string', help='Directory for the GNN output.')
    if 'NT_dir' in io_options:
        group_io.add_option('--NT-dir', dest='NT_dir', type='string', help='Directory for the niche trajectory output.')
    if 'output' in io_options:
        group_io.add_option('-o', '--output', dest='output', type='string', help='Directory for analysis output.')

    # input files
    if 'input' in io_options:
        group_io.add_option(
            '--meta-input',
            dest='meta_input',
            type='string',
            help=
            'Meta data file in csv format. Each row is a cell. The first column should be the cell name with column name Cell_ID. Coordinates (x, y) and sample should be included. Cell type is required for cell-level data.'
        )
    if 'log' in io_options:
        group_io.add_option('-l', '--log', dest='log', type='string', help='Log file.')

    # deprecated options
    if 'NN_dir' in io_options:
        group_io.add_option('--preprocessing-dir',
                            dest='preprocessing_dir',
                            type='string',
                            help='This options will be deprecated from v3.0. Please use --NN-dir instead.')
    if 'NT_dir' in io_options:
        group_io.add_option('--NTScore-dir',
                            dest='NTScore_dir',
                            type='string',
                            help='This options will be deprecated from v3.0. Please use --NT-dir instead.')
    if 'input' in io_options:
        group_io.add_option('-d',
                            '--dataset',
                            dest='dataset',
                            type='string',
                            help='This options will be deprecated from v3.0. Please use --meta-input instead.')

    optparser.add_option_group(group_io)


def validate_io_options(optparser: OptionParser,
                        options: Values,
                        io_options: Optional[List[str]],
                        required: bool = True,
                        overwrite_validation: bool = True) -> None:
    """Validate IO options from a OptParser object.
    :param optparser: OptionParser object.
    :param options: Options object.
    :param io_options: List of I/O options.
    :param required: Required flag.
    :param overwrite_validation: Overwrite validation flag.
    :return: None.
    """
    if io_options is None:
        return

    if 'NN_dir' in io_options:
        if hasattr(options,
                   'preprocessing_dir') and options.preprocessing_dir is not None and (not hasattr(options, 'NN_dir')
                                                                                       or options.NN_dir is None):
            warning('The --preprocessing-dir option will be deprecated from v3.0. Please use --NN-dir instead.')
            options.NN_dir = options.preprocessing_dir
        if required and (not hasattr(options, 'NN_dir') or options.NN_dir is None):
            error('Please provide a directory for niche network outputs.')
            optparser.print_help()
            sys.exit(1)
        if hasattr(options, 'NN_dir') and options.NN_dir is not None and os.path.isdir(options.NN_dir):
            if overwrite_validation:
                warning(f'The directory ({options.NN_dir}) you given already exists. It will be overwritten.')
            else:
                pass
        else:
            info(f'Creating directory: {options.NN_dir}')
            os.makedirs(options.NN_dir, exist_ok=True)

    if 'GNN_dir' in io_options:
        if required and (not hasattr(options, 'GNN_dir') or options.GNN_dir is None):
            error('Please provide a directory for the GNN output.')
            optparser.print_help()
            sys.exit(1)
        if hasattr(options, 'GNN_dir') and options.GNN_dir is not None and os.path.isdir(options.GNN_dir):
            if overwrite_validation:
                warning(f'The directory ({options.GNN_dir}) you given already exists. It will be overwritten.')
            else:
                pass
        else:
            info(f'Creating directory: {options.GNN_dir}')
            os.makedirs(options.GNN_dir, exist_ok=True)

    if 'NT_dir' in io_options:
        if hasattr(options, 'NTScore_dir') and options.NTScore_dir is not None and (not hasattr(options, 'NT_dir')
                                                                                    or options.NT_dir is None):
            warning('The --NTScore-dir option will be deprecated from v3.0. Please use --NT-dir instead.')
            options.NT_dir = options.NTScore_dir
        if required and (not hasattr(options, 'NT_dir') or options.NT_dir is None):
            error('Please provide a directory for the NTScore output.')
            optparser.print_help()
            sys.exit(1)
        if hasattr(options, 'NT_dir') and options.NT_dir is not None and os.path.isdir(options.NT_dir):
            if overwrite_validation:
                warning(f'The directory ({options.NT_dir}) you given already exists. It will be overwritten.')
            else:
                pass
        else:
            info(f'Creating directory: {options.NT_dir}')
            os.makedirs(options.NT_dir, exist_ok=True)

    if 'input' in io_options:

        # meta data
        if not required:
            pass
        else:
            if options.dataset and not options.meta_input:
                warning('The --dataset option will be deprecated from v3.0. Please use --meta-input instead.')
                options.meta_input = options.dataset
            if not hasattr(options, 'meta_input') or options.meta_input is None:
                error('Please provide a meta data file in csv format.')
                optparser.print_help()
                sys.exit(1)
            if not os.path.isfile(options.meta_input):
                error(f'The input file ({options.meta_input}) you given does not exist.')
                optparser.print_help()
                sys.exit(1)
            if not options.meta_input.endswith(('csv', 'csv.gz')):
                error(f'The input file ({options.meta_input}) should be in csv format.')
                optparser.print_help()
                sys.exit(1)

    if 'output' in io_options:  # this is a required option (ONTraC_analysis only)
        if not hasattr(options, 'output') or options.output is None:
            error('Please provide a directory for analysis output.')
            optparser.print_help()
            sys.exit(1)
        if os.path.isdir(options.output):
            if overwrite_validation:
                warning(f'The directory ({options.output}) you given already exists. It will be overwritten.')
            else:
                pass
        else:
            info(f'Creating directory: {options.output}')
            os.makedirs(options.output, exist_ok=True)

    if 'log' in io_options:  # this is a optional option (ONTraC_analysis only)
        if hasattr(options, 'log') and options.log is not None and not os.path.isfile(options.log):
            error(f'Log file: {options.log} you given does not exist.')
            sys.exit(1)


def write_io_options_memo(options: Values, io_options: Optional[List[str]]) -> None:
    """Write IO options to stdout.
    :param options: Options object.
    :param io_options: List of I/O options.
    :return: None.
    """
    if io_options is None:
        return
    info('            -------- I/O options -------             ')
    if 'NN_dir' in io_options and hasattr(options, 'NN_dir') and options.NN_dir is not None:
        info(f'Niche network output directory:  {options.NN_dir}')
    if 'GNN_dir' in io_options and hasattr(options, 'GNN_dir') and options.GNN_dir is not None:
        info(f'GNN output directory:  {options.GNN_dir}')
    if 'NT_dir' in io_options and hasattr(options, 'NT_dir') and options.NT_dir is not None:
        info(f'Niche trajectory output directory:  {options.NT_dir}')
    if 'output' in io_options:
        info(f'Output directory:  {options.output}')
    if 'input' in io_options and hasattr(options, 'meta_input') and options.meta_input is not None:
        info(f'Meta data file:  {options.meta_input}')
    if 'log' in io_options:
        if options.log:
            info(f'Log file:  {options.log}')



import os
import sys
from copy import deepcopy
from pathlib import Path
from typing import Dict, Union

import yaml


def write_version_info() -> None:
    """
    Write version information to stdout.
    :return: None.
    """
    from .. import __version__
    template = f'''##################################################################################

         ▄▄█▀▀██   ▀█▄   ▀█▀ █▀▀██▀▀█                   ▄▄█▀▀▀▄█
        ▄█▀    ██   █▀█   █     ██    ▄▄▄ ▄▄   ▄▄▄▄   ▄█▀     ▀
        ██      ██  █ ▀█▄ █     ██     ██▀ ▀▀ ▀▀ ▄██  ██
        ▀█▄     ██  █   ███     ██     ██     ▄█▀ ██  ▀█▄      ▄
         ▀▀█▄▄▄█▀  ▄█▄   ▀█    ▄██▄   ▄██▄    ▀█▄▄▀█▀  ▀▀█▄▄▄▄▀

                        version: {__version__}

##################################################################################
'''

    sys.stdout.write(template)
    sys.stdout.flush()


def get_rel_params(NN_dir: Union[str, Path], params: Dict) -> Dict:
    """
    Get relative paths for params.
    :param NN_dir: str or Path, NN_dir.
    :param params: Dict, input samples.
    :return: Dict, relative paths.
    """
    rel_params = deepcopy(params)
    for data in rel_params['Data']:
        for k, v in data.items():
            if k == 'Name':
                continue
            data[k] = f'{NN_dir}/{v}'
    return rel_params


def read_yaml_file(yaml_file: str) -> dict:
    """
    Read yaml file.
    :param yaml_file: str, yaml file.
    :return: dict, parameters.
    """
    with open(yaml_file, 'r') as fhd:
        params = yaml.load(fhd, Loader=yaml.FullLoader)
    return params


def count_lines(filename: str) -> int:
    """
    Count lines of a file.
    :param filename: file name.
    :return: number of lines.
    """
    i = 0
    if filename.endswith('.gz'):
        import gzip
        fhd = gzip.open(filename, 'rt')
    else:
        fhd = open(filename, 'r')
    for _ in fhd:
        i += 1
    fhd.close()
    return i


def get_meta_data_file(NN_dir: Union[str, Path]) -> Union[str, Path]:
    """
    Get meta data file.
    :param NN_dir: str or Path, NN_dir.
    :return: str, meta data file.
    """
    # meta data
    meta_data_file = f'{NN_dir}/meta_data.csv.gz'
    if not os.path.isfile(meta_data_file):
        meta_data_file = f'{NN_dir}/meta_data.csv'
    if not os.path.isfile(meta_data_file):
        raise ValueError(
            'meta_data.csv. is required for preprocessing. Copy the input meta_data.csv to the NN_dir may solve this problem.'
        )
    return Path(meta_data_file)


def round_epoch_filter(epoch: int) -> bool:
    """
    Round epoch filter.
    Only round epoch (1, 2, ..., 9, 10, 20, ..., 90, 100, ...) will be saved.
    :param epoch: int.
    :return: bool.
    """

    def _is_power_of_10(n: int) -> bool:
        """
        Check if n is power of 10.
        :param n: int.
        :return: bool.
        """
        num = len(str(n))
        return n % (10**(num - 1)) == 0

    return epoch < 10 or _is_power_of_10(epoch)



from optparse import Values
from typing import Callable, Dict

import numpy as np

from ..data import load_dataset
from ..GNN import evaluate, predict, save_graph_pooling_results, set_seed, train
from ..log import *
from ..model import GNN
from ..niche_net import construct_niche_network, gen_samples_yaml
from ..niche_trajectory import NTScore_table, get_niche_NTScore, load_consolidate_data, niche_to_cell_NTScore
from ..optparser import *
from ..preprocessing.pp_control import preprocessing_gnn, preprocessing_nn
from ..train import GNNBatchTrain
from ..train.inspect_funcs import get_inspect_funcs
from ..utils import get_rel_params, read_yaml_file


def load_parameters(opt_validate_func: Callable, prepare_optparser_func: Callable) -> Values:
    """
    Load parameters.
    :param opt_validate_func: validate function.
    :param prepare_optparser_func: prepare optparser function.
    :return: options.
    """
    options = opt_validate_func(prepare_optparser_func())

    return options


def niche_network_construct(options: Values) -> None:
    """
    Niche network construct process.
    :param options: options.
    :return: None.
    """

    info('------------- Niche network construct --------------- ')

    # load input information
    meta_data_df, ct_coding = preprocessing_nn(meta_input=options.meta_input, NN_dir=options.NN_dir)

    # construct niche network
    construct_niche_network(meta_data_df=meta_data_df,
                            ct_coding=ct_coding,
                            save_dir=options.NN_dir,
                            n_neighbors=options.n_neighbors,
                            n_local=options.n_local)

    # generate samples.yaml to indicate file paths for each sample
    gen_samples_yaml(meta_data_df=meta_data_df, save_dir=options.NN_dir)
    info('------------ Niche network construct end ------------ ')


def gnn(options: Values) -> None:
    """
    GNN training and prediction process.
    :param options: options.
    :return: None.
    """

    info('------------------------ GNN ------------------------ ')
    # load data
    dataset, sample_loader, meta_data_df = preprocessing_gnn(NN_dir=options.NN_dir, batch_size=options.batch_size)
    # random seed
    set_seed(seed=options.seed)
    # device
    device = options.device
    # build model
    input_feats = dataset.num_features
    model = GNN(input_feats=input_feats,
                hidden_feats=options.hidden_feats,
                k=options.k,
                n_gcn_layers=options.n_gcn_layers,
                exponent=options.beta)
    # train
    loss_weight_args: Dict[str, float] = {
        key: value
        for key, value in options.__dict__.items() if key.endswith('loss_weight')
    }
    batch_train = train(nn_model=model,
                        BatchTrain=GNNBatchTrain,
                        sample_loader=sample_loader,
                        device=device,
                        max_epochs=options.epochs,
                        max_patience=options.patience,
                        min_delta=options.min_delta,
                        min_epochs=options.min_epochs,
                        lr=options.lr,
                        save_dir=options.GNN_dir,
                        inspect_funcs=get_inspect_funcs(),
                        **loss_weight_args)
    # evaluate
    evaluate(batch_train=batch_train)
    # predict
    consolidate_s_array, _ = predict(output_dir=options.GNN_dir, batch_train=batch_train, dataset=dataset)
    # save results
    if consolidate_s_array is not None:
        save_graph_pooling_results(meta_data_df=meta_data_df,
                                   dataset=dataset,
                                   rel_params=get_rel_params(NN_dir=options.NN_dir,
                                                             params=read_yaml_file(f'{options.NN_dir}/samples.yaml')),
                                   consolidate_s_array=consolidate_s_array,
                                   output_dir=options.GNN_dir)
    info('--------------------- GNN end ---------------------- ')


def niche_trajectory_construct(options: Values) -> None:
    """
    Pseudotime calculateion process.
    :param options: options.
    :return: None.
    """

    info('----------------- Niche trajectory ------------------ ')
    consolidate_s_array, consolidate_out_adj_array = load_consolidate_data(GNN_dir=options.GNN_dir)

    params = read_yaml_file(f'{options.NN_dir}/samples.yaml')
    rel_params = get_rel_params(NN_dir=options.NN_dir, params=params)
    dataset = load_dataset(NN_dir=options.NN_dir)

    niche_cluster_score, niche_level_NTScore = get_niche_NTScore(
        trajectory_construct_method=options.trajectory_construct,
        niche_cluster_loading=consolidate_s_array,
        niche_adj_matrix=consolidate_out_adj_array)
    cell_level_NTScore, all_niche_level_NTScore_dict, all_cell_level_NTScore_dict = niche_to_cell_NTScore(
        dataset=dataset, rel_params=rel_params, niche_level_NTScore=niche_level_NTScore)

    np.savetxt(fname=f'{options.NT_dir}/niche_cluster_score.csv.gz', X=niche_cluster_score, delimiter=',')
    np.savetxt(fname=f'{options.NT_dir}/niche_NTScore.csv.gz', X=niche_level_NTScore, delimiter=',')
    np.savetxt(fname=f'{options.NT_dir}/cell_NTScore.csv.gz', X=cell_level_NTScore, delimiter=',')

    NTScore_table(save_dir=options.NT_dir,
                  rel_params=rel_params,
                  all_niche_level_NTScore_dict=all_niche_level_NTScore_dict,
                  all_cell_level_NTScore_dict=all_cell_level_NTScore_dict)
    info('--------------- Niche trajectory end ---------------- ')
